/* Radix-2 Fast Fourier Transform, third shot
 * Demonstration code by Emil Mikulic, 2001.
 * http://dmr.ath.cx/sound/fft/
 *
 * Feel free to do whatever you like with this code.
 * Feel free to credit me.
 */

#include <math.h>
#include <stdio.h>
#include <sndfile.h> /* http://www.zip.com.au/~erikd/libsndfile/ */

typedef double val;
#ifndef PI
# ifdef M_PI
#  define PI M_PI
# else
/* overkill */
#  define PI 3.141592653589793238462643383279502884197169399375105820974
# endif
#endif

#define POWER 16
#define SIZE (1<<POWER)

val i_re[SIZE], i_im[SIZE], out_re[SIZE], out_im[SIZE];
SF_INFO sfinfo;
SNDFILE *sndfile;

/* reverse bits according to POWER */
int br(const int i)
{
	int j, k = 0;

	for (j=0; j<POWER; j++)
	if ((i >> j) & 1)
		k = k | (1<<(POWER-j-1));

	return k;
}



/* note: N must be a power of 2 or whacked-out things will happen */
void fft(val *in_re, val *in_im)
{
	int blocks, points, i;

	for (blocks=SIZE/2,points=2; blocks>=1; blocks/=2,points*=2)
	for (i=0; i<blocks; i++)
	{
		int j, offset = points * i;

		/* "butterfly" odd elements */
		for (j=0; j<points/2; j++)
		{
			val t_re, t_im, top_r, top_i, btm_r, btm_i, theta;
			theta = -2.0 * M_PI *
				(val)j / (val)(points);

			t_re = cos(theta);
			t_im = sin(theta);

			top_r = in_re[j+offset];
			top_i = in_im[j+offset];

			btm_r = in_re[j+points/2+offset]*t_re -
				in_im[j+points/2+offset]*t_im;

			btm_i = in_re[j+points/2+offset]*t_im +
				in_im[j+points/2+offset]*t_re;			

			in_re[j+offset] = top_r + btm_r;
			in_im[j+offset] = top_i + btm_i;

			in_re[j+points/2+offset] = top_r - btm_r;
			in_im[j+points/2+offset] = top_i - btm_i;
		}
	}
}



int main()
{
	int i;
	val *tmp;

	sndfile = sf_open_read("sample.wav", &sfinfo);

	/* Informational */
	fprintf(stderr, "Opened sample --\n\n"
		"samplerate:  %d\n"
		"samples:     %d\n"
		"channels:    %d\n"
		"pcmbitwidth: %d\n\n",
		sfinfo.samplerate, sfinfo.samples,
		sfinfo.channels, sfinfo.pcmbitwidth);

	if (sfinfo.channels != 1) {
		fprintf(stderr, "Need mono input.\n");
		sf_close(sndfile);
		return 1;
	}

	/* Allocate memory */
	tmp = (val*)malloc(sizeof(val) * sfinfo.samples);

	if (!tmp) {
		fprintf(stderr, "Out of memory!\n");
		sf_close(sndfile);
		return 1;
	}

	/* Fill up input */
	if (sf_read_double(sndfile, tmp, sfinfo.samples, 1) !=
		sfinfo.samples) {
		fprintf(stderr, "Failed to read %d samples.\n",
			sfinfo.samples);
		sf_close(sndfile);
		return 1;
	}
	sf_close(sndfile);

	/* Sort into bit-reversed order */
	for (i=0; i<SIZE; i++)
	{
		int pos = br(i);

		i_re[pos] = tmp[i];
		i_im[pos] = 0;
	}
	free(tmp);

	fft(i_re, i_im);

	/* Output (and scaling) */
	for (i=0; i<SIZE/2; i++)
	{
		out_re[i] = i_re[i] / (val)SIZE;
		out_im[i] = i_im[i] / (val)SIZE;

		printf("%f Hz: %f\n",
			(double)i*(val)sfinfo.samplerate/(val)SIZE,
			sqrt(out_re[i]*out_re[i] + out_im[i]*out_im[i]) );
	}

	return 0;
}
